# Import das bibliotecas.

# Biblioteca de logging
import logging 

# Biblioteca de aprendizado de máquina
import torch 

# Bibliotecas próprias
from textotransformer.mensurador.medidas import distanciaEuclidiana, distanciaManhattan, similaridadeCoseno
from textotransformer.mensurador.mensuradorenum import PalavrasRelevantes
from textotransformer.modelo.modeloenum import LISTATIPOCAMADA_NOME, EmbeddingsCamadas, EstrategiasPooling 
from textotransformer.util.utilconstantes import OUTPUTS, OUTPUTS_HIDDEN_STATES
from textotransformer.util.utiltexto import encontrarIndiceSubLista  

logger = logging.getLogger(__name__)

class Mensurador:

    ''' 
    Realiza mensurações em textos.
     
    Parâmetros:
    `modelo_args` - Parâmetros do modelo de linguagem.
    `transformer_model` - Modelo de linguagem carregado.
    `pln` - Processador de linguagem natural.
    ''' 

    # Construtor da classe
    def __init__(self, modelo_args, transformer_model, pln):
    
        # Parâmetros do modelo
        self.model_args = modelo_args
    
        # Recupera o objeto do transformer.
        self.transformer_model = transformer_model
    
        # Recupera o modelo.
        self.model = transformer_model.get_auto_model()
    
        # Recupera o tokenizador.     
        self.tokenizer = transformer_model.get_tokenizer()
        
        # Recupera a classe PLN
        self.pln = pln
        
        logger.info("Classe Mensurador carregada: {}.".format(modelo_args))
      
    # ============================
    def getEmbeddingsTodasCamadas(self, texto):    
        '''   
        Retorna os embeddings de todas as camadas de um texto.
        
        Parâmetros:
        `texto` - Um texto a ser recuperado os embeddings do BERT.
        
        Retorno:
        `textoTokenizado` - Texto tokenizado.
        `input_ids` - Input ids do texto.
        `attention_mask` - Máscara de atenção do texto
        `token_type_ids` - Token types ids do texto.
        `outputs` - Embeddings do texto.
        '''

        # Texto tokenizado
        textoTokenizado =  self.transformer_model.getTextoTokenizado(texto)

        #print('O texto (', texto, ') tem tamanho = ', len(textoTokenizado), ' = ', textoTokenizado)

        # Recupera a quantidade tokens do texto tokenizado.
        qtdeTokens = len(textoTokenizado)

        #tokeniza o texto e retorna os tensores.
        dicCodificado = self.tokenizer.encode_plus(
                                                texto, # Texto a ser codificado.
                                                add_special_tokens=True, # Adiciona os tokens especiais '[CLS]' e '[SEP]'
                                                max_length=qtdeTokens, # Define o tamanho máximo para preencheer ou truncar.
                                                truncation=True, # Trunca o texto por max_length
                                                padding='max_length', # Preenche o texto até max_length
                                                return_attention_mask=True, # Constrói a máscara de atenção.
                                                return_tensors='pt' # Retorna os dados como tensores pytorch.
                                                )
        
        # Ids dos tokens de entrada mapeados em seus índices do vocabuário.
        input_ids = dicCodificado['input_ids']

        # Máscara de atenção de cada um dos tokens como pertencentes à sentença '1'.
        attention_mask = dicCodificado['attention_mask']

        # Recupera os tensores dos segmentos.
        token_type_ids = dicCodificado['token_type_ids']

        # Roda o texto através do BERT, e coleta todos os estados ocultos produzidos.
        # das 12 camadas. 
        with torch.no_grad():

            # Passe para a frente, calcule as previsões outputs.     
            outputs = self.model(input_ids=input_ids, 
                                 attention_mask=attention_mask)

            # A avaliação do modelo retorna um número de diferentes objetos com base em
            # como é configurado na chamada do método `from_pretrained` anterior. Nesse caso,
            # porque definimos `output_hidden_states = True`, o terceiro item será o
            # estados ocultos(hidden_states) de todas as camadas. Veja a documentação para mais detalhes:
            # https://huggingface.co/transformers/model_doc/bert.html#bertmodel

            # Retorno de model quando ´output_hidden_states=True´ é setado:    
            # outputs[0] = last_hidden_state, outputs[1] = pooler_output, outputs[2] = hidden_states
            # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
            
            # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        return textoTokenizado, input_ids, attention_mask, token_type_ids, outputs

    # ============================
    # Cria um buffer com os embeddings de sentenças para economizar memória no processamento.
    buffer_embeddings = {}
    
    def getEmbeddingsTodasCamadasBuffer(self, S):
        '''
        Retorna os embeddings de uma sentença de um buffer ou do modelo..
        '''
        
        # Se está no dicionário retorna o embedding
        if S in self.buffer_embeddings:
            return self.buffer_embeddings.get(S)
        else:
            # Gera o embedding
            totalCamada = self.getEmbeddingsTodasCamadas(S)
            self.buffer_embeddings.update({S: totalCamada})
            return totalCamada

    # ============================
    def limpaBufferEmbedding(self):
        '''
        Esvazia o buffer de embeddings das sentenças.
        '''
        
        self.buffer_embeddings.clear(self)

    # ============================
    def getEmbeddingPrimeiraCamada(self, sentencaEmbedding):
        '''
        Retorna os embeddings da primeira camada.
        '''
        
        # Cada elemento do vetor sentencaEmbeddinging é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
      
        # Retorna todas a primeira(-1) camada
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        resultado = sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][0]
        # Retorno: (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        #print('resultado=',resultado.size())

        return resultado

    # ============================
    def getEmbeddingPenultimaCamada(self, sentencaEmbedding):
        '''
        Retorna os embeddings da penúltima camada.
        '''
        
        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
      
        # Retorna todas a penúltima(-2) camada
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        resultado = sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][-2]
        # Retorno: (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        #print('resultado=',resultado.size())

        return resultado

    # ============================
    def getEmbeddingUltimaCamada(self, sentencaEmbedding):
        '''
        Retorna os embeddings da última camada.
        '''
        
        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
      
        # Retorna todas a última(-1) camada
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        resultado = sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][-1]
        # Retorno: (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        #print('resultado=',resultado.size())
      
        return resultado    

    # ============================
    def getEmbeddingSoma4UltimasCamadas(self, sentencaEmbedding):
        '''
        Retorna os embeddings da soma das 4 últimas camadas.
        '''
        
        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
      
        # Retorna todas as 4 últimas camadas
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        embeddingCamadas = sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][-4:]
        # Retorno: List das camadas(4) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  

        # Usa o método `stack` para criar uma nova dimensão no tensor 
        # com a concateção dos tensores dos embeddings.        
        # Entrada: List das camadas(4) (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        resultadoStack = torch.stack(embeddingCamadas, dim=0)
        # Retorno: <4> x <1(lote)> x <qtde_tokens> x <768 ou 1024>
        #print('resultadoStack=',resultadoStack.size())
      
        # Realiza a soma dos embeddings de todos os tokens para as camadas
        # Entrada: <4> x <1(lote)> x <qtde_tokens> x <768 ou 1024>
        resultado = torch.sum(resultadoStack, dim=0)
        # Saida: <1(lote)> x <qtde_tokens> x <768 ou 1024>
        #print('resultado=',resultado.size())

        return resultado

    # ============================
    def getEmbeddingConcat4UltimasCamadas(self, sentencaEmbedding):
        '''
        Retorna os embeddings da concatenação das 4 últimas camadas.
        '''
        
        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
        
        # Cria uma lista com os tensores a serem concatenados
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        # Lista com os tensores a serem concatenados
        listaConcat = []
        # Percorre os 4 últimos
        for i in [-1, -2, -3, -4]:
            # Concatena da lista
            listaConcat.append(sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][i])
        # Retorno: Entrada: List das camadas(4) (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        #print('listaConcat=',len(listaConcat))

        # Realiza a concatenação dos embeddings de todos as camadas
        # Retorno: Entrada: List das camadas(4) (<1(lote)> x <qtde_tokens> <768 ou 1024>)  
        resultado = torch.cat(listaConcat, dim=-1)
        # Retorno: Entrada: (<1(lote)> x <qtde_tokens> <3072 ou 4096>)  
        # print('resultado=',resultado.size())
      
        return resultado   

    # ============================
    def getEmbeddingSomaTodasAsCamada(self, sentencaEmbedding):
        '''
        Retorna os embeddings da soma de todas as camadas.
        '''
        
        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
      
        # Retorna todas as camadas descontando a primeira(0)
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        embeddingCamadas = sentencaEmbedding[OUTPUTS][OUTPUTS_HIDDEN_STATES][1:]
        # Retorno: List das camadas(12 ou 24) (<1(lote)> x <qtde_tokens> <768 ou 1024>)  

        # Usa o método `stack` para criar uma nova dimensão no tensor 
        # com a concateção dos tensores dos embeddings.        
        # Entrada: List das camadas(12 ou 24) (<1(lote)> x <qtde_tokens> x <768 ou 1024>)  
        resultadoStack = torch.stack(embeddingCamadas, dim=0)
        # Retorno: <12 ou 24> x <1(lote)> x <qtde_tokens> x <768 ou 1024>
        #print('resultadoStack=',resultadoStack.size())
      
        # Realiza a soma dos embeddings de todos os tokens para as camadas
        # Entrada: <12 ou 24> x <1(lote)> x <qtde_tokens> x <768 ou 1024>
        resultado = torch.sum(resultadoStack, dim=0)
        # Saida: <1(lote)> x <qtde_tokens> x <768 ou 1024>
        # print('resultado=',resultado.size())
      
        return resultado

    # ============================
    def getResultadoEmbeddings(self, sentencaEmbedding, camada):
        '''
        Retorna o resultado da operação sobre os embeddings das camadas de acordo com tipo de camada especificada.
        
        Parâmetros:
        `sentencaEmbedding` - Embeddings da stentença.
        `camada` - Camada dos embeddings.
        '''

        # Cada elemento do vetor sentencaEmbedding é formado por:  
        # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
        # hidden_states é uma lista python, e cada elemento um tensor pytorch no formado <lote> x <qtde_tokens> x <768 ou 1024>.
        #[4]outpus e [2]hidden_states 
        #[OUTPUTS]outpus e [OUTPUTS_HIDDEN_STATES]hidden_states      
        # Entrada: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> x <768 ou 1024>) 

        resultadoEmbeddingCamadas = None
      
        if camada == EmbeddingsCamadas.PRIMEIRA_CAMADA.value[0]:
            resultadoEmbeddingCamadas = self.getEmbeddingPrimeiraCamada(sentencaEmbedding)
            #print('resultadoEmbeddingCamadas1=',resultadoEmbeddingCamadas.size())
        else:
            if camada == EmbeddingsCamadas.PENULTIMA_CAMADA.value[0]:
                resultadoEmbeddingCamadas = self.getEmbeddingPenultimaCamada(sentencaEmbedding)
                #print('resultadoEmbeddingCamadas1=',resultadoEmbeddingCamadas.size())
            else:                
                if camada == EmbeddingsCamadas.ULTIMA_CAMADA.value[0]:
                    resultadoEmbeddingCamadas = self.getEmbeddingUltimaCamada(sentencaEmbedding)
                    #print('resultadoEmbeddingCamadas2=',resultadoEmbeddingCamadas.size())
                else:
                    if camada == EmbeddingsCamadas.SOMA_4_ULTIMAS_CAMADAS.value[0]:
                        resultadoEmbeddingCamadas = self.getEmbeddingSoma4UltimasCamadas(sentencaEmbedding)
                        #print('resultadoEmbeddingCamadas3=',resultadoEmbeddingCamadas.size())
                    else:                        
                        if camada == EmbeddingsCamadas.CONCAT_4_ULTIMAS_CAMADAS.value[0]:
                            resultadoEmbeddingCamadas = self.getEmbeddingConcat4UltimasCamadas(sentencaEmbedding)
                            #print('resultadoEmbeddingCamadas4=',resultadoEmbeddingCamadas.size())
                        else:
                            if camada == EmbeddingsCamadas.TODAS_AS_CAMADAS.value[0]:                            
                                resultadoEmbeddingCamadas = self.getEmbeddingSomaTodasAsCamada(sentencaEmbedding)
                                #print('resultadoEmbeddingCamadas5=',resultadoEmbeddingCamadas.size())
                                # Retorno: <1> x <qtde_tokens> x <768 ou 1024>
                            else:
                                logger.info("Nenhuma seleção da camada dos embeddings foi especificada.")
          
        # Verifica se a primeira dimensão é igual 1 para remover a dimensão de lote 'batches'
        # Entrada: <1> x <qtde_tokens> x <768 ou 1024>
        if resultadoEmbeddingCamadas.shape[0] == 1:
            # Remove a dimensão 0 caso seja de tamanho 1.
            # Usa o método 'squeeze' para remover a primeira dimensão(0) pois possui tamanho 1
            # Entrada: <1> x <qtde_tokens> x <768 ou 1024>
            resultadoEmbeddingCamadas = torch.squeeze(resultadoEmbeddingCamadas, dim=0)     
        #print('resultadoEmbeddingCamadas2=', resultadoEmbeddingCamadas.size())    
        # Retorno: <qtde_tokens> x <768 ou 1024>
      
        # Retorna o resultados dos embeddings dos tokens da sentença  
        return resultadoEmbeddingCamadas

    # ============================
    def getMedidasSentencasEmbeddingMEAN(self, embeddingSi, embeddingSj):
        '''
        Retorna as medidas de duas sentenças Si e Sj utilizando a estratégia MEAN.
        
        Parâmetros:
        `embeddingSi` - Embeddings da primeira sentença.
        `embeddingSj` - Embeddings da segunda sentença.
        
        Retorno:
        `Scos` - Similaridade do cosseno - usando a média dos embeddings Si e Sj das camadas especificadas.
        `Seuc` - Distância euclidiana - usando a média dos embeddings Si e Sj das camadas especificadas.
        `Sman` - Distância de manhattan - usando a média dos embeddings Si e Sj das camadas especificadas.
        '''

        #print('embeddingSi=', embeddingSi.shape) 
        #print('embeddingSj=', embeddingSj.shape)
      
        # As operações de subtração(sub), mul(multiplicação/produto), soma(sum), cosseno(similaridade), euclediana(diferença) e manhattan(diferença)
        # Necessitam que os embeddings tenha a mesmo número de dimensões.
      
        # Calcula a média dos embeddings para os tokens de Si, removendo a primeira dimensão.
        # Entrada: <qtde_tokens> x <768 ou 1024>  
        mediaEmbeddingSi = torch.mean(embeddingSi, dim=0)    
        # Retorno: <768 ou 1024>
        #print('mediaCamadasSi=', mediaCamadasSi.shape)
      
        # Calcula a média dos embeddings para os tokens de Sj, removendo a primeira dimensão.
        # Entrada: <qtde_tokens> x <768 ou 1024>  
        mediaEmbeddingSj = torch.mean(embeddingSj, dim=0)    
        # Retorno: <768 ou 1024>
        #print('mediaCamadasSj=', mediaCamadasSj.shape)
        
        # Similaridade do cosseno entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Scos = similaridadeCoseno(mediaEmbeddingSi, mediaEmbeddingSj)
        # Retorno: Número real

        # Distância euclidiana entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Seuc = distanciaEuclidiana(mediaEmbeddingSi, mediaEmbeddingSj)
        # Retorno: Número real

        # Distância de manhattan entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Sman = distanciaManhattan(mediaEmbeddingSi, mediaEmbeddingSj)
        # Retorno: Número real

        # Retorno das medidas das sentenças  
        return mediaEmbeddingSi, mediaEmbeddingSj, Scos, Seuc, Sman

    # ============================
    def getMedidasSentencasEmbeddingMAX(self, embeddingSi, embeddingSj):
        '''
        Retorna as medidas de duas sentenças Si e Sj utilizando a estratégia MAX.
        
        Parâmetros:
        `embeddingSi` - Embeddings da primeira sentença.
        `embeddingSj` - Embeddings da segunda sentença.
           
        Retorno:
        `Scos` - Similaridade do cosseno - usando o maior dos embeddings Si e Sj das camadas especificadas.
        `Seuc` - Distância euclidiana - usando o maior dos embeddings Si e Sj das camadas especificadas.
        `Sman` - Distância de manhattan - usando o maior dos embeddings Si e Sj das camadas especificadas.
        '''

        #print('embeddingSi=', embeddingSi.shape) 
        #print('embeddingSj=', embeddingSj.shape)

        # As operações de subtração(sub), mul(multiplicação/produto), soma(sum), cosseno(similaridade), euclediana(diferença) e manhattan(diferença)
        # Necessitam que os embeddings tenha a mesmo número de dimensões.

        # Encontra os maiores embeddings os tokens de Si, removendo a primeira dimensão.
        # Entrada: <qtde_tokens> x <768 ou 1024>  
        maiorEmbeddingSi, linha = torch.max(embeddingSi, dim=0)    
        # Retorno: <768 ou 1024>
        #print('maiorEmbeddingSi=', maiorEmbeddingSi.shape)

        # Encontra os maiores embeddings os tokens de Sj, removendo a primeira dimensão.
        # Entrada: <qtde_tokens> x <768 ou 1024>  
        maiorEmbeddingSj, linha = torch.max(embeddingSj, dim=0)    
        # Retorno: <768 ou 1024>
        #print('maiorEmbeddingSj=', maiorEmbeddingSj.shape)

        # Similaridade do cosseno entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Scos = similaridadeCoseno(maiorEmbeddingSi, maiorEmbeddingSj)
        # Retorno: Número real

        # Distância euclidiana entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Seuc = distanciaEuclidiana(maiorEmbeddingSi, maiorEmbeddingSj)
        # Retorno: Número real

        # Distância de manhattan entre os embeddings Si e Sj
        # Entrada: (<768 ou 1024>) x (<768 ou 1024>)
        Sman = distanciaManhattan(maiorEmbeddingSi, maiorEmbeddingSj)
        # Retorno: Número real

        # Retorno das medidas das sentenças
        return maiorEmbeddingSi, maiorEmbeddingSj, Scos, Seuc, Sman

    # ============================
    def getMedidasSentencasEmbedding(self, embeddingSi, embeddingSj):
        '''
        Realiza o cálculo da medida do texto de acordo com a estratégia de pooling(MAX ou MEAN).
        
        Parâmetros:
        `embeddingSi` - Embeddings da primeira sentença.
        `embeddingSj` - Embeddings da segunda sentença.
        `estrategia_pooling` - Estratégia de pooling a ser utilizada.       
        '''

        if self.model_args.estrategia_pooling == EstrategiasPooling.MEAN.value:
            return self.getMedidasSentencasEmbeddingMEAN(embeddingSi, embeddingSj)
        else:
            if self.model_args.estrategia_pooling == EstrategiasPooling.MAX.value:
                return self.getMedidasSentencasEmbeddingMAX(embeddingSi, embeddingSj)
            else:
                logger.info("Nenhuma seleção da estratégia de pooling foi especificada.")

    # ============================
    def getEmbeddingSentencaEmbeddingTextoALL(self, 
                                              embeddingTexto, 
                                              texto, 
                                              sentenca):
        '''
        Retorna os embeddings de uma sentença com todas as palavras(ALL) a partir dos embeddings do texto.
        
        '''
            
        # Tokeniza o texto
        textoTokenizado =  self.transformer_model.getTextoTokenizado(texto)
        #print(textoTokenizado)

        # Tokeniza a sentença
        sentencaTokenizada =  self.transformer_model.getTextoTokenizado(sentenca)
        #print(sentencaTokenizada)
        # Remove os tokens de início e fim da sentença
        sentencaTokenizada.remove('[CLS]')
        sentencaTokenizada.remove('[SEP]')    
        #print(len(sentencaTokenizada))

        # Localiza os índices dos tokens da sentença no texto
        inicio, fim = encontrarIndiceSubLista(textoTokenizado, sentencaTokenizada)
        #print(inicio,fim) 

        # Recupera os embeddings dos tokens da sentença a partir dos embeddings do texto
        embeddingSentenca = embeddingTexto[inicio:fim + 1]
        #print('embeddingSentenca=', embeddingSentenca.shape)

        # Retorna o embedding da sentença no texto
        return embeddingSentenca

    # ============================
    def getEmbeddingSentencaEmbeddingTextoCLEAN(self, 
                                                embeddingTexto, 
                                                texto, 
                                                sentenca):
        '''
        Retorna os embeddings de uma sentença sem stopwords(CLEAN) a partir dos embeddings do texto.
        '''
          
        # Tokeniza o texto
        textoTokenizado =  self.transformer_model.getTextoTokenizado(texto)  
        #print(textoTokenizado)

        # Remove as stopword da sentença
        sentencaSemStopWord = self.pln.removeStopWord(sentenca)

        # Tokeniza a sentença sem stopword
        sentencaTokenizadaSemStopWord =  self.transformer_model.getTextoTokenizado(sentencaSemStopWord)
        #print(sentencaTokenizadaSemStopWord)

        # Remove os tokens de início e fim da sentença
        sentencaTokenizadaSemStopWord.remove('[CLS]')
        sentencaTokenizadaSemStopWord.remove('[SEP]')    
        #print(len(sentencaTokenizadaSemStopWord))

        # Tokeniza a sentença
        sentencaTokenizada =  self.transformer_model.getTextoTokenizado(sentenca)

        # Remove os tokens de início e fim da sentença
        sentencaTokenizada.remove('[CLS]')
        sentencaTokenizada.remove('[SEP]')  
        #print(sentencaTokenizada)
        #print(len(sentencaTokenizada))

        # Localiza os índices dos tokens da sentença no texto
        inicio, fim = encontrarIndiceSubLista(textoTokenizado, sentencaTokenizada)
        #print('Sentença inicia em:', inicio, 'até', fim) 

        # Recupera os embeddings dos tokens da sentença a partir dos embeddings do texto
        embeddingSentenca = embeddingTexto[inicio:fim + 1]
        #print('embeddingSentenca=', embeddingSentenca.shape)

        # Lista com os tensores selecionados
        listaTokensSelecionados = []
        # Localizar os embeddings dos tokens da sentença tokenizada sem stop word na sentença 
        # Procura somente no intervalo da sentença
        for i, tokenSentenca in enumerate(sentencaTokenizada):
            for tokenSentencaSemStopWord in sentencaTokenizadaSemStopWord: 
                # Se o token da sentença é igual ao token da sentença sem stopword    
                if tokenSentenca == tokenSentencaSemStopWord:
                    listaTokensSelecionados.append(embeddingSentenca[i:i + 1])

        embeddingSentencaSemStopWord = None

        if len(listaTokensSelecionados) != 0:
            # Concatena os vetores da lista pela dimensão 0
            embeddingSentencaSemStopWord = torch.cat(listaTokensSelecionados, dim=0)
            #print("embeddingSentencaSemStopWord:",embeddingSentencaSemStopWord.shape)

        # Retorna o embedding da sentença no texto
        return embeddingSentencaSemStopWord

    # ============================
    def getEmbeddingSentencaEmbeddingTextoNOUN(self, 
                                               embeddingTexto, 
                                               texto, 
                                               sentenca):
        '''
        Retorna os embeddings de uma sentença somente com as palavras relevantes(NOUN) de um tipo a partir dos embeddings do texto.
        '''

        # Tokeniza o texto
        textoTokenizado =  self.transformer_model.getTextoTokenizado(texto)  
        #print(textoTokenizado)

        # Retorna as palavras relevantes da sentença do tipo especificado
        sentencaSomenteRelevante = self.pln.retornaPalavraRelevante(sentenca, self.model_args.palavra_relevante)

        # Tokeniza a sentença 
        sentencaTokenizadaSomenteRelevante =  self.transformer_model.getTextoTokenizado(sentencaSomenteRelevante)

        # Remove os tokens de início e fim da sentença
        sentencaTokenizadaSomenteRelevante.remove('[CLS]')
        sentencaTokenizadaSomenteRelevante.remove('[SEP]')  
        #print(sentencaTokenizadaSomenteRelevante)
        #print(len(sentencaTokenizadaSomenteRelevante))

        # Tokeniza a sentença
        sentencaTokenizada =  self.transformer_model.getTextoTokenizado(sentenca)

        # Remove os tokens de início e fim da sentença
        sentencaTokenizada.remove('[CLS]')
        sentencaTokenizada.remove('[SEP]')  
        #print(sentencaTokenizada)
        #print(len(sentencaTokenizada))

        # Localiza os índices dos tokens da sentença no texto
        inicio, fim = encontrarIndiceSubLista(textoTokenizado, sentencaTokenizada)
        #print('Sentença inicia em:', inicio, 'até', fim) 

        # Recupera os embeddings dos tokens da sentença a partir dos embeddings do texto
        embeddingSentenca = embeddingTexto[inicio:fim + 1]
        #print('embeddingSentenca=', embeddingSentenca.shape)

        # Lista com os tensores selecionados
        listaTokensSelecionados = []
        # Localizar os embeddings dos tokens da sentença tokenizada sem stop word na sentença 
        # Procura somente no intervalo da sentença
        for i, tokenSentenca in enumerate(sentencaTokenizada):
            for tokenSentencaSomenteRelevante in sentencaTokenizadaSomenteRelevante: 
                if tokenSentenca == tokenSentencaSomenteRelevante:        
                    listaTokensSelecionados.append(embeddingSentenca[i:i + 1])

        embeddingSentencaComSubstantivo = None

        if len(listaTokensSelecionados) != 0:
            # Concatena os vetores da lista pela dimensão 0  
            embeddingSentencaComSubstantivo = torch.cat(listaTokensSelecionados, dim=0)
            #print("embeddingSentencaComSubstantivo:",embeddingSentencaComSubstantivo.shape)

        # Retorna o embedding da sentença do texto
        return embeddingSentencaComSubstantivo

    # ============================
    def getEmbeddingSentencaEmbeddingTexto(self, 
                                           embeddingTexto, 
                                           texto, 
                                           sentenca):
        '''
        Retorna os embeddings de uma sentença considerando a relevância das palavras (ALL, CLEAN ou NOUN) a partir dos embeddings do texto.    
        '''

        if self.model_args.palavra_relevante == PalavrasRelevantes.ALL.value:
            return self.getEmbeddingSentencaEmbeddingTextoALL(embeddingTexto, texto, sentenca)
        else:
            if self.model_args.palavra_relevante == PalavrasRelevantes.CLEAN.value:                
                return self.getEmbeddingSentencaEmbeddingTextoCLEAN(embeddingTexto, texto, sentenca)
            else:
                if self.model_args.palavra_relevante == PalavrasRelevantes.NOUN.value:
                    return self.getEmbeddingSentencaEmbeddingTextoNOUN(embeddingTexto, texto, sentenca)
                else:
                    logger.info("Nenhuma estratégia de relevância de palavras foi especificada.") 

    # ============================
    def getMedidasComparacaoTexto(self, 
                                  texto, 
                                  camada, 
                                  tipoTexto='p'):
        '''
        Retorna as medidas de coerência do texto.
        Considera somente sentenças com pelo menos uma palavra.
        Estratégia de pooling padrão é MEAN(0).
        Palavra relavante padrão é ALL(0).
        '''

        # Quantidade de sentenças no texto
        n = len(texto)
        
        # Divisor da quantidade de textos
        divisor = n - 1

        # Texto é uma lista com as sentenças
        #print('camada=',camada)
        #print('Texto=', texto)

        # Junta a lista de sentenças em um texto(string)
        stringTexto = ' '.join(texto)

        # Envia o texto ao MCL e recupera os embeddings de todas as camadas
        # Se for o texto original pega do buffer para evitar a repetição
        if tipoTexto == 'o':
            # Retorna os embeddings de todas as camadas do texto
            # O embedding possui os seguintes valores        
            # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
            totalCamadasTexto = self.getEmbeddingsTodasCamadasBuffer(stringTexto)      
            # Retorno: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> <768 ou 1024>) 
        else:
            # Retorna os embeddings de todas as camadas do texto
            # O embedding possui os seguintes valores        
            # 0-texto_tokenizado, 1-input_ids, 2-attention_mask, 3-token_type_ids, 4-outputs(0=last_hidden_state,1=pooler_output,2=hidden_states)
            totalCamadasTexto = self.getEmbeddingsTodasCamadas(stringTexto)      
            # Retorno: List das camadas(13 ou 25) (<1(lote)> x <qtde_tokens> <768 ou 1024>) 

        # Recupera os embeddings dos tokens das camadas especificadas de acordo com a estratégia especificada para camada  
        embeddingTexto = self.getResultadoEmbeddings(totalCamadasTexto, camada=camada)
        #print('embeddingTexto=', embeddingTexto.shape)

        # Acumuladores das medidas entre as sentenças  
        somaScos = 0
        somaSeuc = 0
        somaSman = 0

        # Seleciona os pares de sentença a serem avaliados
        posSi = 0
        posSj = posSi + 1

        #Enquanto o indíce da sentneça posSj(2a sentença) não chegou ao final da quantidade de sentenças
        while posSj <= (n-1):  

            # Seleciona as sentenças do texto  
            Si = texto[posSi]
            Sj = texto[posSj]

            # Recupera os embedding das sentenças Si e Sj do embedding do texto      
            embeddingSi = self.getEmbeddingSentencaEmbeddingTexto(embeddingTexto, stringTexto, Si)
            embeddingSj = self.getEmbeddingSentencaEmbeddingTexto(embeddingTexto, stringTexto, Sj)

            # Verifica se os embeddings sentenças estão preenchidos
            if embeddingSi != None and embeddingSj != None:

                # Recupera as medidas entre Si e Sj     
                ajustadoEmbeddingSi, ajustadoEmbeddingSj, Scos, Seuc, Sman = self.getMedidasSentencasEmbedding(embeddingSi, embeddingSj)

                # Acumula as medidas
                somaScos = somaScos + Scos
                somaSeuc = somaSeuc + Seuc
                somaSman = somaSman + Sman

                    # avança para o próximo par de sentenças
                posSi = posSj
                posSj = posSj + 1
            else:
                # Reduz um da quantidade de sentenças pois uma delas está vazia
                divisor = divisor - 1
                # Se embeddingSi igual a None avanca pos1 e pos2
                if embeddingSi == None:
                    # Avança a posição da sentença posSi para a posSj
                    posSi = posSj
                    # Avança para a próxima sentença de posSj
                    posSj = posSj + 1        
                else:          
                    # Se embeddingSj = None avança somente posJ para a próxima sentença
                    if embeddingSj == None:
                        posSj = posSj + 1

        # Calcula a medida 
        Ccos = 0
        Ceuc = 0
        Cman = 0

        if divisor != 0:
            Ccos = float(somaScos) / float(divisor)
            Ceuc = float(somaSeuc) / float(divisor)
            Cman = float(somaSman) / float(divisor)

        return Ccos, Ceuc, Cman
   
    # ============================
    def comparaMedidasCamadasSentencas(self, Si, Sj, camada):
        '''
        Facilita a exibição dos valores de comparação de duas orações.
        '''
      
        # Recupera os embeddings da sentença Si e sentença Sj e suas medidas
        embeddingSi, embeddingSj, Scos, Seuc, Sman = self.getMedidasCamadasSentencas(Si, Sj, camada)

        logger.info('  ->Mostra comparação da ' + camada[LISTATIPOCAMADA_NOME] + ' camada(s)')    
        logger.info('   Cosseno(SixSj)     = %.8f' % Scos)
        logger.info('   Euclidiana(SixSj)  = %.8f' % Seuc)
        logger.info('   Manhattan(SixSj)   = %.8f' % Sman)